unit dp_parser;

// Dynamic Pascal, Copyright Lars Middendorf

interface
uses classes,dp_scanner,Contnrs,sysutils,dp_code;

type


TIdentList=array of string;

TTokenInfo=record
 Token:TTokenType;
 TokenString:string;
 SourceLine,SourcePos:integer;
end;

TExpression=class;

TExpressionList=array of TExpression;

TExpression=class
 info:TTokenInfo;
 nodelist:TObjectList;
 subnodes:array[0..1] of TExpression;
 paramlist:array of TExpression;
 symbolinfo:TSymbolInfo;
 typeinfo:TTypeInfo;
 expectedtype:TTypeInfo;
 unary:boolean;
 cgen:TCodeGenerator;
 precedence:integer;
 read,allowtypecast:boolean;
 procedure SyntaxError(const msg:string);
 procedure IncompatibleTypes(t,t2:TTypeInfo);
 function IsConst:boolean;
 function IsLValue:boolean;
 function CalcDepth:integer;
 procedure SetExpressionType(t:TBasetype);
 function ReadInteger:TExpression;
 function ReadFloat:TExpression;
 function ReadString:TExpression;
 function ReadIdent:TExpression;
 function ReadNil:TExpression;
 procedure CheckSwapNodes;
 function ReadOperatorAdd:TExpression;
 function ReadOperatorSub:TExpression;
 function ReadOperatorMul:TExpression;
 function ReadOperatorIDiv:TExpression;
 function ReadOperatorMod:TExpression;
 function ReadOperatorDiv:TExpression;
 function ReadRecordOperator:TExpression;
 function ReadArrayOperator:TExpression;
 function ReadOperatorEqual:TExpression;
 function ReadOperatorNot:TExpression;
 function ReadOperatorNeg:TExpression;
 function ReadOperatorRel:TExpression;
 function ReadOperatorLogic:TExpression;
 function ReadExpression(check:boolean=true):TExpression;
 function CheckFuncCall:TExpression;
 function ReadFuncCall:TExpression;
 procedure WriteIdentAddr;
 procedure WriteRecordAddr;
 procedure WriteArrayAddr;
 procedure GenerateLValueCode;
 procedure WriteInteger;
 procedure WriteFloat;
 procedure WriteString;
 procedure WriteIdent;
 procedure WriteNil;
 procedure WriteOperatorAdd;
 procedure WriteOperatorSub;
 procedure WriteOperatorMult;
 procedure WriteOperatorDiv;
 procedure WriteOperatorMod;
 procedure WriteOperatorEqual;
 procedure WriteOperatorNotEqual;
 procedure WriteOperatorNot;
 procedure WriteOperatorNeg;
 procedure WriteOperatorGreater;
 procedure WriteOperatorGreaterEqual;
 procedure WriteOperatorLess;
 procedure WriteOperatorLessEqual;
 procedure WriteOperatorAnd;
 procedure WriteOperatorOr;
 procedure WriteOperatorXor;
 procedure WriteRecord;
 procedure WriteArray;
 function AddResultParam(ti:TTypeInfo):TExpression;
 procedure WriteFuncCall;
 procedure PushParam(param:TSymbolInfo);
 procedure WriteParamList(ti:TTypeInfo);
 procedure PopParamRegister(ti:TTypeInfo;r:TRegister);
 procedure GenerateCode;
 constructor create(cg:TCodeGenerator;nl:TObjectList);
end;

TOnLoadUnitEvent=function(const name:string):TUnitInfo of object;
TOnLoadUnitFunc=function(const name:string):TStream;

TParser=class
 sc:TScanner;
 nodelist:TObjectList;
 interfacesec:boolean;
 BreakOnEqual:boolean;
 unitinfo:TUnitInfo;
 cclass:TTypeInfo;
 vistype:TFieldVisibility;
 cgen:TCodeGenerator;
 expstack:TStack;
 exptree:TExpression;
 exitjmp:array of integer;
 breakjmp:array of integer;
 contjmp:array of integer;
 loopdepth:integer;
 OnLoadUnit:TOnLoadUnitEvent;
 OnLoadUnitFunc:TOnLoadUnitFunc;
procedure CheckToken(token:TTokenSet);
function ReadTokenInfo:TTokenInfo;
procedure AcceptToken(token:TTokentype);overload;
procedure AcceptToken;overload;
function ParseIdent:string;
function ParseIdentList(endtoken:TTokenSet):TIdentList;
function GetOperatorPrecedence(exp:TExpression):integer;
function CompareOperator(exp1,exp2:TExpression):boolean;
procedure PushExpression(e:TExpression);
function PopExpression:TExpression;
function ParseType:TTypeInfo;
function ParseParamType:TTypeInfo;
function ParseIntegerExpression:Integer;
function ParseVarType:TTypeInfo;
function ParseMethodType:TMethodType;
procedure ParseEndBlock;
procedure ParseFieldDecl;
procedure ParseMethodDecl(method:TSymbolInfo=nil);
procedure ParseParamList(mt:TSymbolInfo);
procedure CheckParamList(mt:TSymbolInfo);
procedure ParseClassDecl(const name:string);
procedure ParseEnumDecl(const name:string);
procedure ParseConstBlock;
procedure ParseTypeBlock;
function LoadUnit(const name:string):TUnitInfo;
procedure ParseInterfaceUses;
procedure ParseInterfaceSection;
function ParseMethodImpl:TSymbolInfo;
procedure ParseVarDeclList;
procedure ParseVarBlock;
procedure AddBinaryOperator(e:TExpression);
procedure AddUnaryOperator(e:TExpression);
procedure AddOperator(e:TExpression);
function ParseExpression_r:TExpression;
function ParseExpression:TExpression;
procedure AssignExpression(exp1,exp2:TExpression);
procedure ParseExpressionStatement;
procedure ParseIfStatement;
procedure ParseWhileStatement;
procedure ParseForStatement;
procedure ParseRepeatStatement;
procedure ParseStatement;
procedure ParseExitStatement;
procedure ParseBreakStatement;
procedure ParseContinueStatement;
procedure ParseCodeBlock;
procedure ParseMethod(mt:TSymbolInfo);
procedure ParseImplementationSection;
function ParseComplexIdent:string;
function Parse:TUnitInfo;
constructor create(st:TStream);
destructor destroy;override;
end;

procedure IncompatibleTypes(t,t2:TTypeInfo);

function CompileUnit(st:TStream):TUnitInfo;overload;
function CompileUnit(const source:string):TUnitInfo;overload;


var
 OnLoadUnit:TOnLoadUnitFunc;

implementation

procedure IncompatibleTypes(t,t2:TTypeInfo);
begin
SyntaxError('Incompatible type: "'+t.Name+'" and "'+t2.Name+'"');
end;

// TExpression

procedure TExpression.SyntaxError(const msg:string);
begin
 raise ESyntaxError.Create('Line '+inttostr(info.SourceLine)+' :'+msg);
end;

procedure TExpression.IncompatibleTypes(t,t2:TTypeInfo);
begin
SyntaxError('Incompatible type: "'+t.Name+'" and "'+t2.Name+'"');
end;

function TExpression.IsConst:boolean;
begin
result:=(info.token in [tkinteger,tkfloat,tkstring,tknil]) or
        ((info.token=tkaddress) and (subnodes[0].info.token=tkident));
end;

function TExpression.IsLValue:boolean;
begin
result:=info.token=tkident;
if (symbolinfo<>nil) and (symbolinfo.ParamType=ptconst) then result:=false;
if info.Token=tkpoint then
begin
 if subnodes[0].typeinfo.BaseType=btclass then
 begin
  result:=true;
 end
 else
 begin
 if subnodes[0].IsLValue then result:=true;
 end;
end;
if info.Token=tkarrayopen then
begin
 if subnodes[0].IsLValue then result:=true;
end;
end;

function TExpression.CalcDepth:integer;
var
 a,b:integer;
begin
if info.token in [tkident,tknil,tkinteger,tkfloat,tkstring] then
 result:=1
else
if assigned(subnodes[0]) and assigned(subnodes[1]) then
begin
 a:=subnodes[0].CalcDepth+1;
 b:=subnodes[1].CalcDepth+1;
 if a>b then result:=b else result:=a;
end
else
if assigned(subnodes[0]) then
 result:=subnodes[0].CalcDepth
else
 result:=0;
end;

procedure TExpression.SetExpressionType(t:TBasetype);
begin
case t of
 btinteger:typeinfo:=System_Integer;
 btboolean:typeinfo:=System_Boolean;
 btsingle:typeinfo:=System_Single;
 btstring:typeinfo:=System_String;
end;
if t=btsingle then
begin
 if (subnodes[0]<>nil) and (subnodes[0].info.token=tkinteger) then subnodes[0].info.token:=tkfloat;
 if (subnodes[1]<>nil) and (subnodes[1].info.token=tkinteger) then subnodes[1].info.token:=tkfloat;
 if (subnodes[0]<>nil) then subnodes[0].expectedtype:=System_Single;
 if (subnodes[1]<>nil) then subnodes[1].expectedtype:=System_Single; 
end;
end;

function TExpression.ReadInteger:TExpression;
begin
result:=self;
typeinfo:=System_Integer;
end;

function TExpression.ReadFloat:TExpression;
begin
result:=self;
typeinfo:=System_Single;
end;

function TExpression.ReadString:TExpression;
begin
result:=self;
typeinfo:=System_String;
end;

function TExpression.ReadIdent:TExpression;
var
 node:TExpression;
begin
result:=self;

symbolinfo:=cgen.FindSymbol(info.tokenstring);

if (symbolinfo=nil) and (cgen.context<>nil) then
begin
 symbolinfo:=cgen.Context.FindField(info.TokenString,cgen.Context,cgen.Context.UnitInfo);
 if symbolinfo<>nil then
 begin
  node:=TExpression.create(cgen,nodelist);
  node.subnodes[0]:=TExpression.create(cgen,nodelist);
  node.subnodes[0].info.Token:=tkident;
  node.subnodes[0].info.TokenString:='Self';
  node.subnodes[1]:=self;
  node.info.Token:=tkpoint;
  result:=node.ReadRecordOperator;
 end;
end;

if symbolinfo=nil then
begin
typeinfo:=cgen.FindType(info.TokenString);
if assigned(typeinfo) then
begin
 if allowtypecast then
 begin
  if (typeinfo.BaseType in [btclass,btclassref]) then
  begin
   info.Token:=tktypecast;
  end;
 end
 else
 begin
  if typeinfo.BaseType=btclass then
   typeinfo:=typeinfo.ClassRefTypeInfo;
 end;
end
else
 typeinfo:=nil;
end;

if (symbolinfo=nil) and (typeinfo=nil) then
begin
 SyntaxError('Undeclared Identifier: '+info.tokenstring);
end;
if symbolinfo<>nil then
if sftrueconst in symbolinfo.flags then
begin
typeinfo:=symbolinfo.typeinfo;
  info.tokenstring:=symbolinfo.ConstData;
case typeinfo.basetype of
 btinteger,btenum:info.token:=tkinteger;
 btboolean:info.Token:=tkinteger;
 btsingle:info.token:=tkfloat;
 btstring:info.token:=tkstring;
end;
end;

if typeinfo=nil then
begin
typeinfo:=symbolinfo.typeinfo;
end;

end;

function TExpression.ReadNil:TExpression;
begin
result:=self;
result.typeinfo:=System_Integer;
end;

procedure TExpression.CheckSwapNodes;
var
 e:TExpression;
begin
if subnodes[1].CalcDepth>subnodes[0].CalcDepth then
begin
e:=subnodes[1];
subnodes[1]:=subnodes[0];
subnodes[0]:=e;
end;
end;


function TExpression.ReadOperatorAdd:TExpression;
var
 t0,t1:TTypeInfo;
begin
result:=self;
subnodes[0]:=subnodes[0].ReadExpression;
subnodes[1]:=subnodes[1].ReadExpression;
t0:=subnodes[0].typeinfo;
t1:=subnodes[1].typeinfo;
if t0.IsInteger and t1.IsInteger then
begin
 SetExpressionType(btinteger);
 if (subnodes[0].info.token=tkinteger) and (subnodes[1].info.token=tkinteger) then
 begin
 info.tokenstring:=inttostr(strtoint64(subnodes[0].info.tokenstring)+strtoint64(subnodes[1].info.tokenstring));
 info.token:=tkinteger;
 end;
end
else
if t0.IsNumeric and t1.IsNumeric then
begin
 if (subnodes[0].info.token in [tkinteger,tkfloat]) and (subnodes[1].info.token in [tkinteger,tkfloat]) then
 begin
 info.tokenstring:=floattostr(strtofloat(subnodes[0].info.tokenstring)+strtofloat(subnodes[1].info.tokenstring));
 info.token:=tkinteger;
 end;
 SetExpressionType(btsingle);
end
else
if t0.IsString and t1.IsString then
begin
 if (subnodes[0].info.token=tkstring) and (subnodes[1].info.token=tkstring) then
 begin
 info.tokenstring:=subnodes[0].info.tokenstring+subnodes[1].info.tokenstring;
 info.token:=tkstring;
 end;

 SetExpressionType(btstring)
end
else
 SyntaxError('Incompatible types');
end;

function TExpression.ReadOperatorSub:TExpression;
var
 t0,t1:TTypeInfo;
begin
result:=self;
subnodes[0]:=subnodes[0].ReadExpression;
subnodes[1]:=subnodes[1].ReadExpression;
t0:=subnodes[0].typeinfo;
t1:=subnodes[1].typeinfo;
if t0.IsInteger and t1.IsInteger then
begin
 SetExpressionType(btinteger);
end
else
if t0.IsNumeric or t1.IsNumeric then
 SetExpressionType(btsingle)
else
 SyntaxError('Incompatible types');
end;

function TExpression.ReadOperatorMul:TExpression;
var
 t0,t1:TTypeInfo;
begin
result:=self;
subnodes[0]:=subnodes[0].ReadExpression;
subnodes[1]:=subnodes[1].ReadExpression;
t0:=subnodes[0].typeinfo;
t1:=subnodes[1].typeinfo;
if t0.IsInteger and t1.IsInteger then
begin
 if subnodes[0].IsConst and subnodes[1].IsConst then
 begin
  info.tokenstring:=inttostr(strtoint(subnodes[0].info.tokenstring)*strtoint(subnodes[1].info.tokenstring));
  info.token:=tkinteger;
 end;
 SetExpressionType(btinteger);
end
else
if t0.IsNumeric or t1.IsNumeric then
 SetExpressionType(btsingle)
else
 SyntaxError('Incompatible types');
 
end;

function TExpression.ReadOperatorIDiv:TExpression;
var
 t0,t1:TTypeInfo;
begin
result:=self;
// Momentan knnen vorzeichen- und nicht vorzeichenbehaftete Typen nicht dividiert werden
subnodes[0]:=subnodes[0].ReadExpression;
subnodes[1]:=subnodes[1].ReadExpression;
t0:=subnodes[0].typeinfo;
t1:=subnodes[1].typeinfo;
if t0.IsInteger and t1.IsInteger then
begin
 SetExpressionType(btinteger);
end
else
 SyntaxError('Incompatible types');
end;

function TExpression.ReadOperatorMod:TExpression;
var
 t0,t1:TTypeInfo;
begin
result:=self;
// Momentan knnen vorzeichen- und nicht vorzeichenbehaftete Typen nicht dividiert werden
subnodes[0]:=subnodes[0].ReadExpression;
subnodes[1]:=subnodes[1].ReadExpression;
t0:=subnodes[0].typeinfo;
t1:=subnodes[1].typeinfo;
if t0.IsInteger and t1.IsInteger then
begin
 SetExpressionType(btinteger);
end
else
 SyntaxError('Incompatible types');
end;

function TExpression.ReadOperatorDiv:TExpression;
var
 t0,t1:TTypeInfo;
begin
result:=self;
subnodes[0]:=subnodes[0].ReadExpression;
subnodes[1]:=subnodes[1].ReadExpression;
t0:=subnodes[0].typeinfo;
t1:=subnodes[1].typeinfo;
if t0.IsNumeric and t1.IsNumeric then
begin
 SetExpressionType(btsingle);
end
else
 SyntaxError('Incompatible types');
end;

function TExpression.ReadRecordOperator:TExpression;
var
 field:TSymbolInfo;
begin
result:=self;

try
 subnodes[0].allowtypecast:=allowtypecast;
 subnodes[0]:=subnodes[0].ReadExpression;
except
 info.TokenString:=subnodes[0].info.TokenString+'.'+subnodes[1].info.TokenString;
 info.Token:=tkident;
 subnodes[0]:=nil;
 subnodes[1]:=nil;
 result:=ReadExpression;
 exit;
end;

if not (subnodes[0].typeinfo.basetype in [btrecord,btclass,btclassref]) then
 subnodes[0].SyntaxError('Record type expected');

field:=subnodes[0].typeinfo.FindField(subnodes[1].info.TokenString,cgen.Context,cgen.CurrentFunction.UnitInfo);
if field=nil then
 subnodes[0].SyntaxError('Undeclared identifier: "'+subnodes[1].info.tokenstring+'"');
typeinfo:=field.TypeInfo;
subnodes[1].symbolinfo:=field;
subnodes[1].typeinfo:=field.TypeInfo;
symbolinfo:=field;
end;

function TExpression.ReadArrayOperator:TExpression;
begin
result:=self;
subnodes[0]:=subnodes[0].ReadExpression;
if not (subnodes[0].typeinfo.BaseType in [btarray,btopenarray]) then
 subnodes[0].SyntaxError('Array type expected');
subnodes[1]:=subnodes[1].ReadExpression;
if not subnodes[1].typeinfo.IsInteger then
 subnodes[1].SyntaxError('Integer expression expected');
typeinfo:=subnodes[0].typeinfo.BaseTypePtr;
end;

function TExpression.ReadOperatorEqual;
begin
result:=self;
subnodes[0]:=subnodes[0].ReadExpression;
subnodes[1]:=subnodes[1].ReadExpression;
if not subnodes[0].typeinfo.Equals(subnodes[1].typeinfo) then
begin
 if subnodes[0].typeinfo.IsInteger and subnodes[1].typeinfo.IsFloat then
 begin
  SetExpressionType(btsingle);
 end
 else
 if subnodes[1].typeinfo.IsInteger and subnodes[0].typeinfo.IsFloat then
 begin
  SetExpressionType(btsingle);
 end
 else
  IncompatibleTypes(subnodes[0].typeinfo,subnodes[1].typeinfo);
end;
if not (subnodes[0].typeinfo.BaseType in [btinteger,btboolean,btenum,btsingle,btstring,btclass,btclassref]) then
 SyntaxError('Operator not applicable to this operand type');

 SetExpressionType(btboolean);
end;

function TExpression.ReadOperatorNot:TExpression;
begin
result:=self;
subnodes[0]:=subnodes[0].ReadExpression;
if subnodes[0].typeinfo.BaseType<>btboolean then
 SyntaxError('Operator not applicable to this operand type');
typeinfo:=System_Boolean;
end;

function TExpression.ReadOperatorNeg:TExpression;
begin
result:=self;
subnodes[0]:=subnodes[0].ReadExpression;
if not subnodes[0].typeinfo.IsNumeric then
 SyntaxError('Operator not applicable to this operand type');
typeinfo:=subnodes[0].typeinfo;
end;

function TExpression.ReadOperatorRel:TExpression;
begin
result:=self;
subnodes[0]:=subnodes[0].ReadExpression;
subnodes[1]:=subnodes[1].ReadExpression;
if not (subnodes[0].typeinfo.IsNumeric and subnodes[1].typeinfo.IsNumeric) then
 SyntaxError('Operator not applicable to this operand type');
if subnodes[0].typeinfo.IsFloat or subnodes[1].typeinfo.IsFloat then
 SetExpressionType(btsingle);
typeinfo:=System_Boolean;
end;

function TExpression.ReadOperatorLogic:TExpression;
begin
result:=self;
subnodes[0]:=subnodes[0].ReadExpression;
subnodes[1]:=subnodes[1].ReadExpression;
if (subnodes[0].typeinfo=System_Integer) and (subnodes[1].typeinfo=System_Integer) then
begin
 result.typeinfo:=System_Integer;
end
else
if (subnodes[0].typeinfo=System_Boolean) and (subnodes[1].typeinfo=System_Boolean) then
begin
 result.typeinfo:=System_Boolean;
end
else
 SyntaxError('Operator not applicable to this operand type');
end;

function TExpression.CheckFuncCall:TExpression;
begin
result:=self;
if (info.Token<>tkfunccall) and (typeinfo.BaseType=btmethod) then
begin
result:=TExpression.create(cgen,nodelist);
result.subnodes[0]:=self;
result.info.Token:=tkfunccall;
result:=result.ReadFuncCall;
end;
end;

function TExpression.ReadFuncCall:TExpression;
var
 i,j,k:integer;
begin
result:=self;
if read then exit;
read:=true;
subnodes[0].allowtypecast:=true;
subnodes[0]:=subnodes[0].ReadExpression(false);
if subnodes[0].info.Token=tktypecast then
begin
 if length(paramlist)<>1 then SyntaxError('Invalid typecast');
 paramlist[0]:=paramlist[0].ReadExpression;
 if not (paramlist[0].typeinfo.BaseType in [btclass,btclassref]) then SyntaxError('Invalid typecast');
 result:=paramlist[0];
 result.typeinfo:=subnodes[0].typeinfo;
 exit;
end;

if subnodes[0].subnodes[1]<>nil then
begin
 //symbolinfo:=cgen.FindSymbol(subnodes[0].info.TokenString);
 //symbolinfo:=subnodes[0].typeinfo.FindField(subnodes[1].info.TokenString,cgen.Context);
 typeinfo:=subnodes[0].typeinfo;
end
else
begin
 symbolinfo:=cgen.FindSymbol(subnodes[0].info.TokenString);
 if symbolinfo=nil then SyntaxError('Undeclared identifier: "'+subnodes[0].info.TokenString+'"');
 typeinfo:=symbolinfo.TypeInfo;
end;

if typeinfo.Method then j:=1 else j:=0;
if typeinfo.MethodType in [mtconstructor,mtdestructor] then inc(j);
if typeinfo.ResultPtr then k:=TypeInfo.FieldCount-2 else k:=TypeInfo.FieldCount-1;
if length(paramlist)<>k-j+1 then
 SyntaxError('Invalid paramter count');
for i:=j to k do
begin
 paramlist[i-j].expectedtype:=typeinfo.fields[i].TypeInfo;
 paramlist[i-j]:=paramlist[i-j].ReadExpression;
 if sfreference in typeinfo.Fields[i].Flags then
 begin
  if not paramlist[i-j].typeinfo.Equals(typeinfo.fields[i].TypeInfo) then
   paramlist[i-j].IncompatibleTypes(paramlist[i-j].typeinfo,typeinfo.fields[i].TypeInfo);
 end
 else
 begin
  if not paramlist[i-j].typeinfo.CanAssignTo(typeinfo.fields[i].TypeInfo) then
   paramlist[i-j].IncompatibleTypes(paramlist[i-j].typeinfo,typeinfo.fields[i].TypeInfo);
 end;
end;
if typeinfo.MethodType=mtconstructor then
begin
 typeinfo:=subnodes[0].subnodes[0].typeinfo;
 if typeinfo.BaseType=btclassref then typeinfo:=typeinfo.BaseTypePtr;
end
else
 typeinfo:=typeinfo.BaseTypePtr;
if typeinfo=nil then typeinfo:=system_procedure;

end;

function TExpression.ReadExpression(check:boolean=true):TExpression;
begin
result:=self;
case info.token of
 tkident:result:=ReadIdent;
 tkinteger:result:=ReadInteger;
 tkfloat:result:=ReadFloat;
 tkstring:result:=ReadString;
 tknil:result:=ReadNil;
 tkadd:result:=ReadOperatorAdd;
 tksub:if unary then result:=ReadOperatorNeg else result:=ReadOperatorSub;
 tkmult:result:=ReadOperatorMul;
 tkidiv:result:=ReadOperatorIDiv;
 tkmod:result:=ReadOperatorMod;
 tkdiv:result:=ReadOperatorDiv;
 tkpoint:result:=ReadRecordOperator;
 tkarrayopen:result:=ReadArrayOperator;
 tkequal,tknotequal:result:=ReadOperatorEqual;
 tknot:result:=ReadOperatorNot;
 tkgreater,tkless,tkgequal,tklequal:result:=ReadOperatorRel;
 tkand,tkor,tkxor:result:=ReadOperatorLogic;
 tkfunccall:result:=ReadFuncCall;
end;
if check then result:=result.CheckFuncCall;
end;

procedure TExpression.WriteIdentAddr;
begin
cgen.LoadVarAddress(symbolinfo);
end;

procedure TExpression.WriteRecordAddr;
begin
if (subnodes[0].typeinfo.BaseType in [btclass,btclassref])
   and not (subnodes[0].info.Token in [tkpoint,tkarray,tkident]) then
begin
 subnodes[0].GenerateLValueCode;
 cgen.AddFieldAddress(subnodes[1].symbolinfo);

// subnodes[0].GenerateCode;
// cgen.Mov_EDI_EAX;
// cgen.AddEDIAddress(subnodes[1].symbolinfo.Offset);

end
else
begin
 subnodes[0].GenerateLValueCode;
 cgen.AddFieldAddress(subnodes[1].symbolinfo);
end;
end;

procedure TExpression.WriteArrayAddr;
var
 i:integer;
begin
subnodes[0].GenerateLValueCode;
if subnodes[1].IsConst then
begin
 i:=(StrToInt(subnodes[1].info.TokenString)-subnodes[0].typeinfo.Min)*typeinfo.Size;
 cgen.AddEDIAddress(i);
end
else
begin
 cgen.Mov_ESI_EDI;
 subnodes[1].GenerateCode;
 if typeinfo.Min<>0 then cgen.AddIntConst(-subnodes[0].typeinfo.Min);
 cgen.MultIntConst(typeinfo.Size);
 cgen.Mov_EDI_EAX;
 cgen.Add_EDI_ESI;
end;
end;

procedure TExpression.GenerateLValueCode;
begin
case info.Token of
 tkident:WriteIdentAddr;
 tkpoint:WriteRecordAddr;
 tkarrayopen:WriteArrayAddr;
 else GenerateCode;
end;
end;

procedure TExpression.WriteInteger;
begin
if (expectedtype<>nil) and (expectedtype.BaseType=btsingle) then
begin
 WriteFloat;
 typeinfo:=expectedtype;
end  
else
 cgen.LoadConst(StrToInt(info.tokenstring));
end;

procedure TExpression.WriteFloat;
begin
cgen.LoadConst(StrToFloat(info.tokenstring));
end;

procedure TExpression.WriteString;
begin
cgen.LoadConst(info.tokenstring);
end;

procedure TExpression.WriteIdent;
begin
if typeinfo.BaseType=btclassref then
begin
cgen.LoadConst(typeinfo.BaseTypePtr);
end
else
begin
if (expectedtype<>nil) and (expectedtype.BaseType=btsingle) and (typeinfo.BaseType=btInteger) then
begin
 cgen.LoadIntVar2Float(symbolinfo);
 typeinfo:=System_Single;
end
else
 cgen.LoadVar(symbolinfo);
end;
end;

procedure TExpression.WriteNil;
begin
cgen.LoadConst(0);
end;

procedure TExpression.WriteRecord;
begin
GenerateLValueCode;
case typeinfo.BaseType of
 btinteger,btenum,btclass,btclassref:cgen.LoadIntRefEDI;
 btboolean:cgen.LoadByteRefEDI;
 btsingle:cgen.LoadFloatRefEDI;
 else
  cgen.Mov_EDI_EAX;
end;
end;

procedure TExpression.WriteArray;
begin
GenerateLValueCode;
case typeinfo.BaseType of
 btinteger,btenum,btclass,btclassref:cgen.LoadIntRefEDI;
 btboolean:cgen.LoadByteRefEDI;
 btsingle:cgen.LoadFloatRefEDI;
end;
end;

procedure TExpression.WriteOperatorAdd;
var
 s:TSymbolInfo;
begin
case typeinfo.BaseType of
 btinteger:
 begin
  subnodes[0].GenerateCode;
  case subnodes[1].info.Token of
   tkident:cgen.AddIntLocal(subnodes[1].symbolinfo);
   tkinteger:cgen.AddIntConst(strtoint(subnodes[1].info.TokenString));
   else
   begin
    cgen.Mov_EDX_EAX;
    subnodes[1].GenerateCode;
    cgen.AddIntValue;
   end;
  end;
 end;
 btsingle:
 begin
  subnodes[0].GenerateCode;
  case subnodes[1].info.Token of
   tkident:cgen.AddFloatLocal(subnodes[1].symbolinfo);
   tkfloat:cgen.AddFloatConst(strtofloat(subnodes[1].info.TokenString));
   else
   begin
    subnodes[1].GenerateCode;
    if subnodes[1].typeinfo.BaseType=btinteger then cgen.IntToFloat;
    cgen.AddFloatValue;
   end;
  end;
 end;
 btstring:
 begin
  s:=cgen.AllocateTempVar(System_String);
  subnodes[0].GenerateCode;
  cgen.Mov_ESI_EDI;
  subnodes[1].GenerateCode;
  cgen.AddStringValue(s);
 end;
end;
end;

procedure TExpression.WriteOperatorSub;
begin
case typeinfo.BaseType of
 btinteger:
 begin
  case subnodes[1].info.Token of
   tkident:
   begin
    subnodes[0].GenerateCode;
    cgen.SubIntLocal(subnodes[1].symbolinfo);
   end;
   tkinteger:
   begin
    subnodes[0].GenerateCode;
    cgen.SubIntConst(strtoint(subnodes[1].info.TokenString));
   end;
   else
   begin
    subnodes[1].GenerateCode;
    cgen.Mov_EDX_EAX;
    subnodes[0].GenerateCode;
    cgen.SubIntValue;
   end;
  end;
 end;
 btsingle:
 begin
  subnodes[0].GenerateCode;
  case subnodes[1].info.Token of
   tkident:cgen.SubFloatLocal(subnodes[1].symbolinfo);
   tkfloat:cgen.SubFloatConst(strtofloat(subnodes[1].info.TokenString));
   else
   begin
    subnodes[1].GenerateCode;
    if subnodes[1].typeinfo.BaseType=btinteger then cgen.IntToFloat;
    cgen.SubFloatValue;
   end;
  end;
 end;
end;
end;


procedure TExpression.WriteOperatorMult;
begin
case typeinfo.BaseType of
 btinteger:
 begin
  subnodes[0].GenerateCode;
  case subnodes[1].info.Token of
   tkident:cgen.MultIntLocal(subnodes[1].symbolinfo);
   tkinteger:cgen.MultIntConst(strtoint(subnodes[1].info.TokenString));
   else
   begin
    cgen.Mov_EDX_EAX;
    subnodes[1].GenerateCode;
    cgen.MultIntValue;
   end;
  end;
 end;
 btsingle:
 begin
  subnodes[0].GenerateCode;
  case subnodes[1].info.Token of
   tkident:cgen.MultFloatLocal(subnodes[1].symbolinfo);
   tkfloat:cgen.MultFloatConst(strtofloat(subnodes[1].info.TokenString));
   else
   begin
    subnodes[1].GenerateCode;
    if subnodes[1].typeinfo.BaseType=btinteger then cgen.IntToFloat;
    cgen.MultFloatValue;
   end;
  end;
 end;
end;
end;

procedure TExpression.WriteOperatorDiv;
begin
case typeinfo.BaseType of
 btinteger:
 begin
  case subnodes[1].info.Token of
   tkident:
   begin
    subnodes[0].GenerateCode;
    cgen.DivIntLocal(subnodes[1].symbolinfo);
   end;
   tkinteger:
   begin
    subnodes[0].GenerateCode;
    cgen.DivIntConst(strtoint(subnodes[1].info.TokenString));
   end;
   else
   begin
    subnodes[1].GenerateCode;
    cgen.Mov_EDX_EAX;
    subnodes[0].GenerateCode;
    cgen.DivIntValue;
   end;
  end;
 end;
 btsingle:
 begin
  subnodes[0].GenerateCode;
  case subnodes[1].info.Token of
   tkident:cgen.DivFloatLocal(subnodes[1].symbolinfo);
   tkfloat:cgen.DivFloatConst(strtofloat(subnodes[1].info.TokenString));
   else
   begin
    subnodes[1].GenerateCode;
    if subnodes[1].typeinfo.BaseType=btinteger then cgen.IntToFloat;
    cgen.DivFloatValue;
   end;
  end;
 end;
end;
end;

procedure TExpression.WriteOperatorMod;
begin
case typeinfo.BaseType of
 btinteger:
 begin
  case subnodes[1].info.Token of
   tkident:
   begin
    subnodes[0].GenerateCode;
    cgen.ModIntLocal(subnodes[1].symbolinfo);
   end;
   tkinteger:
   begin
    subnodes[0].GenerateCode;
    cgen.ModIntConst(strtoint(subnodes[1].info.TokenString));
   end;
   else
   begin
    subnodes[1].GenerateCode;
    cgen.Mov_EDX_EAX;
    subnodes[0].GenerateCode;
    cgen.ModIntValue;
   end;
  end;
 end;
 btsingle:
 begin
  subnodes[0].GenerateCode;
  case subnodes[1].info.Token of
   tkident:cgen.DivFloatLocal(subnodes[1].symbolinfo);
   tkfloat:cgen.DivFloatConst(strtofloat(subnodes[1].info.TokenString));
   else
   begin
    subnodes[1].GenerateCode;
    if subnodes[1].typeinfo.BaseType=btinteger then cgen.IntToFloat;
    cgen.DivFloatValue;
   end;
  end;
 end;
end;
end;

procedure TExpression.WriteOperatorEqual;
var
 bt:TBaseType;
begin
bt:=subnodes[0].typeinfo.BaseType;
if (subnodes[0].typeinfo.IsFloat<>subnodes[1].typeinfo.IsFloat)
    and (subnodes[0].typeinfo.BaseType in [btinteger,btsingle]) then bt:=btsingle;
case bt of
 btinteger,btenum,btboolean,btclass,btclassref:
  case subnodes[1].info.Token of
   tkident:
   begin
    subnodes[0].GenerateCode;
    cgen.EqualIntLocal(subnodes[1].symbolinfo);
   end;
   tkinteger:
   begin
    subnodes[0].GenerateCode;
    cgen.EqualIntConst(strtoint(subnodes[1].info.TokenString));
   end;
   else
   begin
    subnodes[0].GenerateCode;
    cgen.Mov_EDX_EAX;
    subnodes[1].GenerateCode;
    cgen.EqualIntValue;
   end;
  end;
 btsingle:
  case subnodes[1].info.Token of
   tkident:
   begin
    subnodes[0].GenerateCode;
    cgen.EqualFloatLocal(subnodes[1].symbolinfo);
   end;
   tkinteger:
   begin
    subnodes[0].GenerateCode;
    cgen.EqualFloatConst(strtoint(subnodes[1].info.TokenString));
   end;
   else
   begin
    subnodes[0].GenerateCode;
    subnodes[1].GenerateCode;
    cgen.EqualFloatValue;
   end;
  end;
 btstring:
 begin
  subnodes[0].GenerateCode;
  cgen.Mov_ESI_EDI;
  subnodes[1].GenerateCode;
  cgen.EqualStringValue;
 end;
end;
end;

procedure TExpression.WriteOperatorNotEqual;
var
 bt:TBaseType;
begin
bt:=subnodes[0].typeinfo.BaseType;
if (subnodes[0].typeinfo.IsFloat<>subnodes[1].typeinfo.IsFloat)
    and (subnodes[0].typeinfo.BaseType in [btinteger,btsingle]) then bt:=btsingle;
case bt of
 btinteger,btenum,btboolean,btclass,btclassref:
  case subnodes[1].info.Token of
   tkident:
   begin
    subnodes[0].GenerateCode;
    cgen.NotEqualIntLocal(subnodes[1].symbolinfo);
   end;
   tkinteger:
   begin
    subnodes[0].GenerateCode;
    cgen.NotEqualIntConst(strtoint(subnodes[1].info.TokenString));
   end;
   else
   begin
    subnodes[0].GenerateCode;
    cgen.Mov_EDX_EAX;
    subnodes[1].GenerateCode;
    cgen.NotEqualIntValue;
   end;
  end;
 btsingle:
  case subnodes[1].info.Token of
   tkident:
   begin
    subnodes[0].GenerateCode;
    cgen.NotEqualFloatLocal(subnodes[1].symbolinfo);
   end;
   tkinteger:
   begin
    subnodes[0].GenerateCode;
    cgen.NotEqualFloatConst(strtoint(subnodes[1].info.TokenString));
   end;
   else
   begin
    subnodes[0].GenerateCode;
    subnodes[1].GenerateCode;
    cgen.NotEqualFloatValue;
   end;
  end;
 btstring:
 begin
  subnodes[0].GenerateCode;
  cgen.Mov_ESI_EDI;
  subnodes[1].GenerateCode;
  cgen.NotEqualStringValue;
 end;
end;
end;

procedure TExpression.WriteOperatorNot;
begin
subnodes[0].GenerateCode;
cgen.NotBoolean;
end;

procedure TExpression.WriteOperatorNeg;
begin
subnodes[0].GenerateCode;
case typeinfo.BaseType of
 btinteger:cgen.NegInteger;
 btsingle:cgen.NegFloat;
end;
end;

procedure TExpression.WriteOperatorGreater;
var
 bt:TBaseType;
begin
bt:=subnodes[0].typeinfo.BaseType;
if (subnodes[0].typeinfo.IsFloat<>subnodes[1].typeinfo.IsFloat)
    and (subnodes[0].typeinfo.BaseType in [btinteger,btsingle]) then bt:=btsingle;
case bt of
 btinteger,btenum,btboolean,btclass,btclassref:
  case subnodes[1].info.Token of
   tkident:
   begin
    subnodes[0].GenerateCode;
    cgen.GreaterIntLocal(subnodes[1].symbolinfo);
   end;
   tkinteger:
   begin
    subnodes[0].GenerateCode;
    cgen.GreaterIntConst(strtoint(subnodes[1].info.TokenString));
   end;
   else
   begin
    subnodes[0].GenerateCode;
    cgen.Mov_EDX_EAX;
    subnodes[1].GenerateCode;
    cgen.GreaterIntValue;
   end;
  end;
 btsingle:
  case subnodes[1].info.Token of
   tkident:
   begin
    subnodes[0].GenerateCode;
    cgen.GreaterFloatLocal(subnodes[1].symbolinfo);
   end;
   tkinteger:
   begin
    subnodes[0].GenerateCode;
    cgen.GreaterFloatConst(strtoint(subnodes[1].info.TokenString));
   end;
   else
   begin
    subnodes[0].GenerateCode;
    subnodes[1].GenerateCode;
    cgen.GreaterFloatValue;
   end;
  end;
end;
end;

procedure TExpression.WriteOperatorGreaterEqual;
var
 bt:TBaseType;
begin
bt:=subnodes[0].typeinfo.BaseType;
if (subnodes[0].typeinfo.IsFloat<>subnodes[1].typeinfo.IsFloat)
    and (subnodes[0].typeinfo.BaseType in [btinteger,btsingle]) then bt:=btsingle;
case bt of
 btinteger,btenum,btboolean,btclass,btclassref:
  case subnodes[1].info.Token of
   tkident:
   begin
    subnodes[0].GenerateCode;
    cgen.GreaterEqualIntLocal(subnodes[1].symbolinfo);
   end;
   tkinteger:
   begin
    subnodes[0].GenerateCode;
    cgen.GreaterEqualIntConst(strtoint(subnodes[1].info.TokenString));
   end;
   else
   begin
    subnodes[0].GenerateCode;
    cgen.Mov_EDX_EAX;
    subnodes[1].GenerateCode;
    cgen.GreaterEqualIntValue;
   end;
  end;
 btsingle:
  case subnodes[1].info.Token of
   tkident:
   begin
    subnodes[0].GenerateCode;
    cgen.GreaterEqualFloatLocal(subnodes[1].symbolinfo);
   end;
   tkinteger:
   begin
    subnodes[0].GenerateCode;
    cgen.GreaterEqualFloatConst(strtoint(subnodes[1].info.TokenString));
   end;
   else
   begin
    subnodes[0].GenerateCode;
    subnodes[1].GenerateCode;
    cgen.GreaterEqualFloatValue;
   end;
  end;
end;
end;

procedure TExpression.WriteOperatorLess;
var
 bt:TBaseType;
begin
bt:=subnodes[0].typeinfo.BaseType;
if (subnodes[0].typeinfo.IsFloat<>subnodes[1].typeinfo.IsFloat)
    and (subnodes[0].typeinfo.BaseType in [btinteger,btsingle]) then bt:=btsingle;
case bt of
 btinteger,btenum,btboolean,btclass,btclassref:
  case subnodes[1].info.Token of
   tkident:
   begin
    subnodes[0].GenerateCode;
    cgen.LessIntLocal(subnodes[1].symbolinfo);
   end;
   tkinteger:
   begin
    subnodes[0].GenerateCode;
    cgen.LessIntConst(strtoint(subnodes[1].info.TokenString));
   end;
   else
   begin
    subnodes[0].GenerateCode;
    cgen.Mov_EDX_EAX;
    subnodes[1].GenerateCode;
    cgen.LessIntValue;
   end;
  end;
 btsingle:
  case subnodes[1].info.Token of
   tkident:
   begin
    subnodes[0].GenerateCode;
    cgen.LessFloatLocal(subnodes[1].symbolinfo);
   end;
   tkinteger:
   begin
    subnodes[0].GenerateCode;
    cgen.LessFloatConst(strtoint(subnodes[1].info.TokenString));
   end;
   else
   begin
    subnodes[0].GenerateCode;
    subnodes[1].GenerateCode;
    cgen.LessFloatValue;
   end;
  end;
end;
end;

procedure TExpression.WriteOperatorLessEqual;
var
 bt:TBaseType;
begin
bt:=subnodes[0].typeinfo.BaseType;
if (subnodes[0].typeinfo.IsFloat<>subnodes[1].typeinfo.IsFloat)
    and (subnodes[0].typeinfo.BaseType in [btinteger,btsingle]) then bt:=btsingle;
case bt of
 btinteger,btenum,btboolean,btclass,btclassref:
  case subnodes[1].info.Token of
   tkident:
   begin
    subnodes[0].GenerateCode;
    cgen.LessEqualIntLocal(subnodes[1].symbolinfo);
   end;
   tkinteger:
   begin
    subnodes[0].GenerateCode;
    cgen.LessEqualIntConst(strtoint(subnodes[1].info.TokenString));
   end;
   else
   begin
    subnodes[0].GenerateCode;
    cgen.Mov_EDX_EAX;
    subnodes[1].GenerateCode;
    cgen.LessEqualIntValue;
   end;
  end;
 btsingle:
  case subnodes[1].info.Token of
   tkident:
   begin
    subnodes[0].GenerateCode;
    cgen.LessEqualFloatLocal(subnodes[1].symbolinfo);
   end;
   tkinteger:
   begin
    subnodes[0].GenerateCode;
    cgen.LessEqualFloatConst(strtoint(subnodes[1].info.TokenString));
   end;
   else
   begin
    subnodes[0].GenerateCode;
    subnodes[1].GenerateCode;
    cgen.LessEqualFloatValue;
   end;
  end;
end;
end;

procedure TExpression.WriteOperatorAnd;
begin
case subnodes[1].info.Token of
 tkident:
 begin
  subnodes[0].GenerateCode;
  if typeinfo.BaseType=btboolean then
   cgen.AndBooleanLocal(subnodes[1].symbolinfo)
  else
   cgen.AndIntLocal(subnodes[1].symbolinfo);
 end;
 tkinteger:
 begin
  subnodes[0].GenerateCode;
  cgen.AndConst(strtoint(subnodes[1].info.TokenString));
 end;
 else
 begin
  subnodes[0].GenerateCode;
  cgen.Mov_EDX_EAX;
  subnodes[1].GenerateCode;
  cgen.AndValue;
 end;
end;
end;

procedure TExpression.WriteOperatorOr;
begin
case subnodes[1].info.Token of
 tkident:
 begin
  subnodes[0].GenerateCode;
  if typeinfo.BaseType=btboolean then
   cgen.OrBooleanLocal(subnodes[1].symbolinfo)
  else
   cgen.OrIntLocal(subnodes[1].symbolinfo);
 end;
 tkinteger:
 begin
  subnodes[0].GenerateCode;
  cgen.OrConst(strtoint(subnodes[1].info.TokenString));
 end;
 else
 begin
  subnodes[0].GenerateCode;
  cgen.Mov_EDX_EAX;
  subnodes[1].GenerateCode;
  cgen.OrValue;
 end;
end;
end;

procedure TExpression.WriteOperatorXor;
begin
case subnodes[1].info.Token of
 tkident:
 begin
  subnodes[0].GenerateCode;
  if typeinfo.BaseType=btboolean then
   cgen.XorBooleanLocal(subnodes[1].symbolinfo)
  else
   cgen.XorIntLocal(subnodes[1].symbolinfo);
 end;
 tkinteger:
 begin
  subnodes[0].GenerateCode;
  cgen.XorConst(strtoint(subnodes[1].info.TokenString));
 end;
 else
 begin
  subnodes[0].GenerateCode;
  cgen.Mov_EDX_EAX;
  subnodes[1].GenerateCode;
  cgen.XorValue;
 end;
end;
end;


procedure TExpression.PushParam(param:TSymbolInfo);
begin
if sfreference in param.Flags then
begin
 GenerateLValueCode;
 cgen.PushReg(regedi);
end
else
begin
 case param.TypeInfo.BaseType of
  btinteger,btboolean,btenum,btclass,btclassref:
  begin
   GenerateCode;
   cgen.PushReg(regeax);
  end;
  btsingle:
  begin
   GenerateCode;
   cgen.pushfloat;
  end;
  btrecord,btarray,btstring:
  begin
   GenerateLValueCode;
   cgen.PushReg(regedi);
   //cgen.PushParamBlock(param.TypeInfo.Size);
  end;
 end;
end;
end;

procedure TExpression.WriteParamList(ti:TTypeInfo);
const
 reglist:array[0..3] of TRegister=(regnone,regeax,regecx,regedx);
var
 i,j,r:integer;
begin
if ti.Method then j:=1 else j:=0;
if ti.MethodType in [mtconstructor,mtdestructor] then inc(j);
for r:=0 to 3 do
 for i:=0 to length(paramlist)-1 do
  if ti.fields[i+j].ParamRegister=reglist[r] then
  begin
   paramlist[i].PushParam(ti.fields[i+j]);
  end;

end;

procedure TExpression.PopParamRegister(ti:TTypeInfo;r:TRegister);
var
 i,j:integer;
begin
if ti.Method then j:=1 else j:=0;
if typeinfo.MethodType in [mtconstructor,mtdestructor] then inc(j);
 for i:=0 to length(paramlist)-1 do
  if ti.fields[i+j].ParamRegister=r then
  begin
   cgen.PopReg(r);
   break;
  end;
end;

function TExpression.AddResultParam(ti:TTypeInfo):TExpression;
var
 l:integer;
 r:TSymbolInfo;
begin
r:=cgen.AllocateTempVar(ti);
l:=length(paramlist);
setlength(paramlist,l+1);
paramlist[l]:=TExpression.create(cgen,nodelist);
paramlist[l].info.Token:=tkident;
paramlist[l].symbolinfo:=r;
paramlist[l].typeinfo:=r.TypeInfo;
result:=paramlist[l];
end;

procedure TExpression.WriteFuncCall;
var
 c:integer;
 r:TExpression;
begin
if cgen.edx_count>0 then
 cgen.PushReg(regedx);
c:=cgen.edx_count;
cgen.edx_count:=0;
cgen.SaveStack;
if subnodes[0].subnodes[1]=nil then
begin
if symbolinfo.typeinfo.ResultPtr then r:=AddResultParam(symbolinfo.typeinfo) else r:=nil;
WriteParamList(symbolinfo.TypeInfo);
PopParamRegister(symbolinfo.TypeInfo,regedx);
PopParamRegister(symbolinfo.TypeInfo,regecx);
PopParamRegister(symbolinfo.TypeInfo,regeax);
cgen.CallAddr(@symbolinfo.MethodPtr);
end
else
begin
 if subnodes[0].subnodes[1].typeinfo.ResultPtr then
  r:=AddResultParam(subnodes[0].subnodes[1].typeinfo)
 else
  r:=nil;
 WriteParamList(subnodes[0].subnodes[1].typeinfo);
 subnodes[0].subnodes[0].GenerateCode;
 PopParamRegister(subnodes[0].subnodes[1].typeinfo,regedx);
 PopParamRegister(subnodes[0].subnodes[1].typeinfo,regecx);
 cgen.callMethod(subnodes[0].subnodes[0].typeinfo.BaseType=btclass,subnodes[0].subnodes[1].symbolinfo);
end;
if cgen.edx_count>0 then
 cgen.PopReg(regedx);
cgen.edx_count:=c;
cgen.RestoreStack;
if typeinfo.BaseType=btboolean then cgen.Movzx_EAX_AL; // zur Sicherheit
if assigned(r) then r.GenerateLValueCode;
end;

procedure TExpression.GenerateCode;
begin
case info.Token of
 tkident:WriteIdent;
 tkinteger:WriteInteger;
 tkfloat:WriteFloat;
 tkstring:WriteString;
 tkadd:WriteOperatorAdd;
 tksub:if unary then WriteOperatorNeg else WriteOperatorSub;
 tkmult:WriteOperatorMult;
 tkdiv:WriteOperatorDiv;
 tkidiv:WriteOperatorDiv;
 tkmod:WriteOperatorMod;
 tkequal:WriteOperatorEqual;
 tknotequal:WriteOperatorNotEqual;
 tkgreater:WriteOperatorGreater;
 tkgequal:WriteOperatorGreaterEqual;
 tkless:WriteOperatorLess;
 tklequal:WriteOperatorLessEqual;
 tknot:WriteOperatorNot;
 tkand:WriteOperatorAnd;
 tkor:WriteOperatorOr;
 tkxor:WriteOperatorXor; 
 tkpoint:WriteRecord;
 tkarrayopen:WriteArray;
 tkfunccall:WriteFuncCall;
end;
if typeinfo.IsInteger and (expectedtype<>nil) and (expectedtype.IsFloat) then
begin
 cgen.IntToFloat;
 SetExpressionType(btsingle);
end;
end;

constructor TExpression.create(cg:TCodeGenerator;nl:TObjectList);
begin
inherited create;
nodelist:=nl;
cgen:=cg;
nl.Add(self);
end;

// TParser

procedure TParser.CheckToken(token:TTokenSet);
begin
if not (sc.TokenType in token) then SyntaxError('Invalid Token');
end;

function TParser.ReadTokenInfo:TTokenInfo;
begin
Result.token:=sc.TokenType;
Result.sourceline:=sc.SourceLine;
Result.SourcePos:=sc.SourcePos;
Result.TokenString:=sc.TokenString;
end;

procedure TParser.AcceptToken(token:TTokentype);
begin
CheckToken([token]);
sc.ReadToken;
end;

procedure TParser.AcceptToken;
begin
sc.ReadToken;
end;

function TParser.ParseIdent:string;
begin
result:=sc.TokenString;
AcceptToken(tkident);
end;

function TParser.ParseIdentList(endtoken:TTokenSet):TIdentList;
var
 i,l:integer;
begin
l:=1;
repeat
 setlength(result,l);
 for i:=0 to l-2 do
  if sametext(result[i],sc.TokenString) then Redeclared(sc.TokenString);
 result[l-1]:=ParseIdent;
 inc(l);
 if sc.TokenType in endtoken then break;
  AcceptToken(tkcomma);
until false;
end;

function TParser.GetOperatorPrecedence(exp:TExpression):integer;
begin
result:=-1;
case exp.info.token of
 tkpoint:result:=5;
 tknot,tkaddress,tkfunccall:result:=4;
 tkmult,tkdiv,tkidiv,tkmod,tkand,tkshl,tkshr:result:=3;
 tkadd,tksub,tkor,tkxor:if exp.unary then result:=4 else result:=2;
 tkequal,tknotequal,tklequal,tkless,tkgreater,tkgequal:result:=1;
end;
end;

function TParser.CompareOperator(exp1,exp2:TExpression):boolean;
begin
result:=GetOperatorPrecedence(exp1)>=GetOperatorPrecedence(exp2);
if (exp1.info.token=tkaddress) and (exp2.info.token=tkaddress) then result:=false;
end;

procedure TParser.PushExpression(e:TExpression);
begin
expstack.Push(e);
exptree:=e;
end;

function TParser.PopExpression:TExpression;
begin
if expstack.Count=0 then
 result:=nil
else
 result:=TExpression(expstack.Pop);
end;

function TParser.ParseType:TTypeInfo;
var
 s:string;
begin
s:=ParseComplexIdent;
result:=unitinfo.FindType(s);
if result=nil then
 Undeclared(s);
end;

function TParser.ParseParamType:TTypeInfo;
begin
if sc.TokenType=tkarray then
begin
AcceptToken(tkarray);
AcceptToken(tkof);
result:=ParseType.CreateOpenArray;
end
else
 result:=ParseType;
end;

function TParser.ParseIntegerExpression:Integer;
var
 exp:TExpression;
begin
exp:=ParseExpression;
exp.ReadExpression;
if not exp.IsConst then SyntaxError('Constant expression expected');
if not exp.typeinfo.IsInteger then SyntaxError('Integer expression expected');
result:=strtoint(exp.info.TokenString);
end;

function TParser.ParseVarType:TTypeInfo;
var
 i,l:integer;
 minlist:array of integer;
 maxlist:array of integer;
begin
if sc.TokenType=tkarray then
begin
AcceptToken(tkarray);
AcceptToken(tkarrayopen);
l:=0;
repeat
 setlength(minlist,l+1);
 setlength(maxlist,l+1);
 if l>0 then AcceptToken(tkcomma);
 minlist[l]:=ParseIntegerExpression;
 AcceptToken(tkrange);
 maxlist[l]:=ParseIntegerExpression;
 if maxlist[l]<minlist[l] then
  SyntaxError('Low bound exceeds high bound');
until sc.TokenType=tkarrayclose;
AcceptToken(tkarrayclose);
AcceptToken(tkof);
result:=ParseVarType;
for i:=l to 0 do
 result:=result.CreateArray(minlist[i],maxlist[i]);
end
else
 result:=ParseType;

end;

function TParser.ParseMethodType:TMethodType;
begin
result:=mtprocedure;
CheckToken([tkclass,tkfunction,tkprocedure,tkconstructor,tkdestructor]);
case sc.TokenType of
 tkfunction:result:=mtfunction;
 tkprocedure:result:=mtprocedure;
 tkconstructor:result:=mtconstructor;
 tkdestructor:result:=mtdestructor;
 tkclass:
 begin
  AcceptToken;
  CheckToken([tkprocedure,tkfunction]);
  if sc.TokenType=tkprocedure then result:=mtclassprocedure else result:=mtclassfunction;
 end;
end;
AcceptToken;
end;

procedure TParser.ParseEndBlock;
begin
CheckToken([tkend,tksemicolon]);
if sc.Tokentype=tksemicolon then AcceptToken(tksemicolon);
end;

procedure TParser.ParseFieldDecl;
var
 namelist:TIdentList;
 i:integer;
 t:TTypeInfo;
begin
namelist:=ParseIdentList([tkdblpoint]);
AcceptToken(tkdblpoint);
t:=ParseType;
 for i:=0 to length(namelist)-1 do
 begin
  cclass.AddField(namelist[i],t,vistype);
 end;
ParseEndBlock;
end;

procedure TParser.ParseParamList(mt:TSymbolInfo);
var
 ptype:TParamType;
 namelist:TIdentList;
 t:TTypeInfo;
 i:integer;
begin
namelist:=nil;
AcceptToken(tkopenbracket);
while sc.TokenType<>tkclosebracket do
begin
 case sc.TokenType of
  tkconst:
   begin
    ptype:=ptconst;
    AcceptToken;
   end;
  tkvar:
  begin
   ptype:=ptvar;
   AcceptToken;
  end;
  else
   ptype:=ptvalue;
 end;
 namelist:=ParseIdentList([tkdblpoint]);
 AcceptToken(tkdblpoint);
 t:=ParseParamType;
 for i:=0 to length(namelist)-1 do
 begin
  if sametext(namelist[i],'Self') or sametext(namelist[i],'Result') then
   Redeclared(namelist[i]);
  mt.TypeInfo.AddParam(namelist[i],t,ptype);
 end;
 CheckToken([tksemicolon,tkclosebracket]);
 if sc.TokenType=tksemicolon then
  AcceptToken
 else
  break;
end;
AcceptToken(tkclosebracket);
end;

procedure TParser.CheckParamList(mt:TSymbolInfo);
var
 ptype:TParamType;
 namelist:TIdentList;
 t:TTypeInfo;
 i,j:integer;
 b:boolean;
begin
namelist:=nil;
if mt.TypeInfo.Method then j:=1 else j:=0;
b:=false;
for i:=0 to mt.TypeInfo.FieldCount-1 do
 if not (sfimplicit in mt.TypeInfo.Fields[i].Flags) then
 begin
  b:=true;
  break;
 end; 
if (sc.TokenType=tkdblpoint) and not b then exit;
AcceptToken(tkopenbracket);
while sc.TokenType<>tkclosebracket do
begin
 case sc.TokenType of
  tkconst:
   begin
    ptype:=ptconst;
    AcceptToken;
   end;
  tkvar:
  begin
   ptype:=ptvar;
   AcceptToken;
  end;
  else
   ptype:=ptvalue;
 end;
 namelist:=ParseIdentList([tkdblpoint]);
 AcceptToken(tkdblpoint);
 t:=ParseParamType;
 for i:=0 to length(namelist)-1 do
 begin
  if j>=mt.TypeInfo.FieldCount then
   SyntaxError('Declaration of "'+mt.TypeInfo.Name+'" differs from previous declaration.');  
  if not (mt.TypeInfo.Fields[j].TypeInfo.Equals(t) and (mt.TypeInfo.Fields[j].ParamType=ptype)
          and (sametext(namelist[i],mt.TypeInfo.Fields[j].Name))) then
   SyntaxError('Declaration of "'+mt.TypeInfo.Name+'" differs from previous declaration.');
  inc(j);
 end;
 CheckToken([tksemicolon,tkclosebracket]);
 if sc.TokenType=tksemicolon then
  AcceptToken
 else
  break;
end;
AcceptToken(tkclosebracket);
end;

procedure TParser.ParseMethodDecl(method:TSymbolInfo=nil);
var
 fname:string;
 mt:TMethodType;
 m:TSymbolInfo;
begin
if method=nil then
begin
if (cclass=nil) then
 CheckToken([tkprocedure,tkfunction]);

mt:=ParseMethodType;
fname:=ParseIdent;
if cclass<>nil then
begin
 m:=cclass.AddMethod(fname,mt,vistype);
 if m.TypeInfo.MethodType in [mtclassprocedure,mtclassfunction] then
 begin
  m.TypeInfo.AddParam('Self',cclass.ClassRefTypeInfo,ptvalue,[sfimplicit]);
 end
 else
 begin
  if cclass.BaseType=btclass then
   m.TypeInfo.AddParam('Self',cclass,ptvalue,[sfimplicit])
  else
   m.TypeInfo.AddParam('Self',cclass,ptvar,[sfimplicit]);
 end;
 if m.TypeInfo.MethodType in [mtconstructor,mtdestructor] then
  m.TypeInfo.AddParam('',System_Integer,ptvalue,[sfimplicit]);
end
else
begin
 m:=unitinfo.AddFunction(fname,mt);
end;
end
else
begin
m:=method;
mt:=method.TypeInfo.methodtype;
end;

if sc.TokenType=tkopenbracket then
 ParseParamList(m);
if mt in [mtclassfunction,mtfunction] then
begin
AcceptToken(tkdblpoint);
m.TypeInfo.BaseTypePtr:=ParseType;
if m.TypeInfo.BaseTypePtr.BaseType in [btrecord,btarray,btstring] then
begin
 m.TypeInfo.AddParam('Result',m.TypeInfo.BaseTypePtr,ptvar,[sfimplicit]);
 m.TypeInfo.ResultPtr:=true;
end;
end;
AcceptToken(tksemicolon);
if (cclass<>nil) and (cclass.BaseType=btclass) then
begin
 if sc.TokenType in [tkvirtual,tkoverride,tkmessage] then
 begin
  case sc.TokenType of
   tkvirtual:
   begin
    cclass.MakeVirtual(m);
    AcceptToken;
   end;
   tkoverride:
   begin
    cclass.OverrideMethod(m);
    AcceptToken;
   end;     
   tkmessage:
   begin
    AcceptToken;
    if not m.TypeInfo.CheckMessageFunc then SyntaxError('Invalid message method declaration');
    m.Flags:=m.Flags+[sfmessage];
    m.MessageID:=ParseIntegerExpression;
    if (m.MessageID<0) or (m.MessageID>high(word)) then SyntaxError('Invalid message id');
   end;
  end;
  AcceptToken(tksemicolon);
 end;
end; 
end;

procedure TParser.ParseClassDecl(const name:string);
var
 baseclass:TTypeInfo;
 rec:boolean;
begin
rec:=sc.TokenType=tkrecord;
AcceptToken;
if rec then
begin
if sc.TokenType=tkopenbracket then
begin
  AcceptToken(tkopenbracket);
  baseclass:=ParseType;
  if baseclass.BaseType<>btrecord then
   SyntaxError('Record type expected.');
  AcceptToken(tkclosebracket);
 end
 else
  baseclass:=nil;
end
else
begin
if sc.TokenType=tkopenbracket then
begin
  AcceptToken(tkopenbracket);
  baseclass:=ParseType;
  if baseclass.BaseType<>btclass then
   SyntaxError('Class type expected.');
  AcceptToken(tkclosebracket);
 end
 else
  baseclass:=unitinfo.FindType('TObject');
end;
if unitinfo.IsDeclared(name) then Redeclared(name);
if rec then cclass:=unitinfo.AddRecord(name,baseclass) else cclass:=unitinfo.AddClass(name,baseclass);
CheckToken([tkident,tkprocedure,tkfunction,tkconstructor,tkdestructor,tkclass,tkend,tkprivate,tkprotected,tkpublic,tkpublished]);
vistype:=fspublic;
while (sc.TokenType<>tkend) do
begin
case sc.TokenType of
 tkident:ParseFieldDecl;
 tkprivate:
 begin
  vistype:=fsprivate;
  AcceptToken;
 end;
 tkprotected:
 begin
  vistype:=fsprotected;
  AcceptToken;
 end; 
 tkpublic:
 begin
  vistype:=fspublic;
  AcceptToken;
 end;
 tkpublished:
 begin
  vistype:=fspublished;
  AcceptToken;
 end; 
 tkprocedure,tkfunction,tkconstructor,tkdestructor,tkclass:ParseMethodDecl;
end;
CheckToken([tkident,tkprocedure,tkfunction,tkconstructor,tkdestructor,tkclass,tkend,tkprivate,tkprotected,tkpublic,tkpublished]);
end;
AcceptToken;
AcceptToken(tksemicolon);
cclass:=nil;
end;

procedure TParser.ParseEnumDecl(const name:string);
var
 namelist:TIdentList;
begin
AcceptToken;
namelist:=ParseIdentList([tkclosebracket]);
AcceptToken;
AcceptToken(tksemicolon);
unitinfo.AddEnum(name,namelist);
end;

procedure TParser.ParseTypeBlock;
var
 s:string;
begin
CheckToken([tkident,tkbegin,tktype,tkconst]);
while (sc.TokenType=tkident) do
begin
 s:=ParseIdent;
 AcceptToken(tkequal);
 CheckToken([tkclass,tkrecord,tkopenbracket]);
 case sc.TokenType of
  tkclass,tkrecord:ParseClassDecl(s);
  tkopenbracket:ParseEnumDecl(s);
 end;
end;
end;

procedure TParser.ParseConstBlock;
var
 s:string;
 exp:TExpression;
begin
CheckToken([tkident,tkbegin,tktype,tkconst]);
while (sc.TokenType=tkident) do
begin
 s:=ParseIdent;
 AcceptToken(tkequal);
 exp:=ParseExpression;
 exp:=exp.ReadExpression;
 AcceptToken(tksemicolon);
 if not exp.IsConst then
   SyntaxError('Constant expression expected');
 unitinfo.AddConst(s,exp.info.TokenString,exp.typeinfo);
end;
end;

function TParser.LoadUnit(const name:string):TUnitInfo;
var
 p:TParser;
 st:TStream;
begin
st:=dp_parser.OnLoadUnit(name);
p:=TParser.create(st);
try
 result:=p.Parse;
finally
 p.free;
 st.free;
end;
end;

procedure TParser.ParseInterfaceUses;
var
 s:string;
 u:TUnitInfo;
begin
AcceptToken(tkuses);
repeat
 s:=ParseComplexIdent;
 u:=TUnitInfo.FindUnit(s);
 if u=nil then u:=onLoadUnit(s);
 if u.InterfaceSec then SyntaxError('Circular reference to unit "'+s+'"');
 unitinfo.AddImport(u);
 if sc.tokentype=tksemicolon then break;
 AcceptToken(tkcomma);
until false;
 AcceptToken(tksemicolon);
end;

procedure TParser.ParseInterfaceSection;
begin
unitinfo.InterfaceSec:=True;
AcceptToken(tkinterface);
if sc.TokenType=tkuses then ParseInterfaceUses;

CheckToken([tkimplementation,tktype,tkconst,tkprocedure,tkfunction]);
while (sc.TokenType<>tkimplementation) do
begin
 case sc.TokenType of
  tktype:
  begin
   AcceptToken(tktype);
   ParseTypeBlock;
  end;
  tkconst:
  begin
   AcceptToken(tkconst);
   ParseConstBlock;
  end;
  tkprocedure,tkfunction:ParseMethodDecl;
 end;
CheckToken([tkimplementation,tktype,tkconst,tkprocedure,tkfunction]);
end;
end;

function TParser.ParseMethodImpl:TSymbolInfo;
var
 mt:TMethodType;
 clsname,methodname:string;
 ti,ti2:TTypeInfo;
 nocheck:boolean;
begin
mt:=ParseMethodType;
clsname:=ParseIdent;
nocheck:=false;
if sc.TokenType=tkpoint then
begin
 AcceptToken(tkpoint);
 methodname:=ParseIdent;
 ti:=unitinfo.FindType(clsname);
 if ti=nil then Undeclared(clsname+'.'+methodname);
 result:=ti.FindField(methodname,ti,unitinfo);
end
else
begin
 methodname:=clsname;
 clsname:=unitinfo.Name;
 result:=unitinfo.FindSymbol(methodname);
 if result=nil then
 begin
  result:=unitinfo.AddFunction(methodname,mt);
  ParseMethodDecl(result);
  nocheck:=true;
 end;
end;
 if result=nil then Undeclared(clsname+'.'+methodname);
 if mt<>result.TypeInfo.MethodType then
  SyntaxError('Declaration of "'+clsname+'.'+methodname+'" differs from previous declaration.');
 if result.Implemented then Redeclared(clsname+'.'+methodname);

if not nocheck then
begin
 if sc.TokenType=tksemicolon then
 begin
  AcceptToken;
 end
 else
 begin
  CheckParamList(result);
  if mt in [mtclassfunction,mtfunction] then
  begin
   AcceptToken(tkdblpoint);
   ti2:=ParseType;
   if not ti2.Equals(result.TypeInfo.BaseTypePtr) then
    SyntaxError('Declaration of "'+clsname+'.'+methodname+'" differs from previous declaration.');
  end;
 AcceptToken(tksemicolon);
 end;
 end;
 result.Implemented:=true; 
end;
procedure TParser.ParseVarDeclList;
var
 namelist:TIdentList;
 t:TTypeInfo;
 i:integer;
begin
namelist:=ParseIdentList([tkdblpoint]);
AcceptToken(tkdblpoint);
t:=ParseVarType;
AcceptToken(tksemicolon);
for i:=0 to length(namelist)-1 do
begin
 if cgen.FindLocal(namelist[i])<>-1 then Redeclared(namelist[i]);
 cgen.AddLocal(namelist[i],t);
end;
end;

procedure TParser.ParseVarBlock;
begin
AcceptToken(tkvar);
while sc.TokenType=tkident do
begin
 ParseVarDeclList;
end;
end;

procedure TParser.AddBinaryOperator(e:TExpression);
begin
if expstack.count<2 then
begin
 raise ESyntaxError.Create('Expression expected but '+sc.TokenString+' found');
end;
e.subnodes[1]:=TExpression(expstack.pop);
e.subnodes[0]:=TExpression(expstack.pop);
PushExpression(e);
exptree:=e;
end;

procedure TParser.AddUnaryOperator(e:TExpression);
begin
if expstack.count<1 then
begin
 raise ESyntaxError.Create('Expression expected but '+sc.TokenString+' found');
end;
e.subnodes[0]:=TExpression(expstack.pop);
PushExpression(e);
exptree:=e;
end;

procedure TParser.AddOperator(e:TExpression);
begin
if e.unary then AddUnaryOperator(e) else AddBinaryOperator(e);
e.precedence:=GetOperatorPrecedence(e);
end;


function TParser.ParseExpression_r:TExpression;
const
 endtoken=[tkcomma,tksemicolon,tkassignment,tkrange,tkthen,tkelse,tkwhile,tkfor,tkto,tkdo,tkrepeat,tkuntil];
var
 e:TExpression;
 preoperator:boolean;
 opstack:TStack;
 l:integer;
 funccall:boolean;
begin
CheckToken([tkident,tkinteger,tkfloat,tkstring,tkopenbracket]+unaryop);
preoperator:=true;
opstack:=TStack.Create;
funccall:=false;
repeat
case sc.TokenType of
 tkident:
  begin
   e:=TExpression.create(cgen,nodelist);
   e.info:=ReadTokenInfo;
   PushExpression(e);
   AcceptToken(tkident);
   CheckToken(binaryop+endtoken+[tkarrayopen,tkopenbracket,tkarrayclose,tkclosebracket,tkderef]);
   preoperator:=false;
   funccall:=true;
  end;
 tkinteger,tkstring,tkfloat:
  begin
   e:=TExpression.create(cgen,nodelist);
   e.info:=ReadTokenInfo;
   PushExpression(e);
   AcceptToken(sc.TokenType);
   CheckToken(binaryop+endtoken+[tkarrayclose,tkclosebracket]);
   preoperator:=false;
  funccall:=false;
  end;
  tkadd,tksub,tkmult,tkdiv,tkequal,
  tklequal,tkless,tkgequal,tkgreater,tknotequal,
  tkidiv,tkmod,tkshr,tkshl,tkand,tkor,tkxor:
  begin
   if (sc.TokenType=tkequal) and BreakOnEqual then break;
   e:=TExpression.create(cgen,nodelist);
   e.info:=ReadTokenInfo;
   e.unary:=preoperator or (sc.TokenType in [tknot,tkaddress]);
   AcceptToken(e.info.token);
   while (opstack.count>0) and CompareOperator(TExpression(opstack.Peek),e) do
   begin
    AddOperator(TExpression(opstack.Pop));
   end;
   opstack.push(e);
   if preoperator then
   begin
    CheckToken(unaryop+[tkident,tkinteger,tkfloat,tkstring,tkopenbracket]-[tkadd,tksub]);
    preoperator:=false;
   end
   else
   begin
    CheckToken(unaryop+[tkident,tkinteger,tkfloat,tkstring,tkopenbracket]);
    preoperator:=true;
   end;
  funccall:=false;
  end;
 tkpoint:
 begin
  e:=TExpression.create(cgen,nodelist);
  e.info:=ReadTokenInfo;
  e.unary:=false;
  AcceptToken(e.info.token);
  while (opstack.count>0) and CompareOperator(TExpression(opstack.Peek),e) do
  begin
   AddOperator(TExpression(opstack.Pop));
  end;
  opstack.push(e);
  CheckToken([tkident]);
  preoperator:=false;
  funccall:=false;
 end;
 tkarrayopen:
 begin
  e:=TExpression.create(cgen,nodelist);
  e.info:=ReadTokenInfo;
  AcceptToken(tkarrayopen);
  e.unary:=true;
  e.subnodes[1]:=ParseExpression;
  AddUnaryOperator(e);
  if sc.TokenType=tkcomma then
  begin
   AcceptToken(tkcomma);
   sc.TokenType:=tkarrayopen;
  end
  else
  begin
   AcceptToken(tkarrayclose);
   CheckToken(binaryop+endtoken+[tkarrayopen,tkarrayclose,tkclosebracket,tkderef]);
  end;
  preoperator:=false;
  funccall:=false;
 end;
 tkopenbracket:
 begin
  if funccall then
  begin
   e:=TExpression.create(cgen,nodelist);
   e.info:=ReadTokenInfo;
   e.info.token:=tkfunccall;
   e.unary:=true;
   AcceptToken(tkopenbracket);
   while sc.TokenType<>tkclosebracket do
   begin
    l:=length(e.paramlist);
    setlength(e.paramlist,l+1);
    e.paramlist[l]:=ParseExpression;
    CheckToken([tkcomma,tkclosebracket]);
    if sc.TokenType=tkcomma then AcceptToken(tkcomma);
   end;
   while (opstack.count>0) and CompareOperator(TExpression(opstack.Peek),e) do
   begin
    AddOperator(TExpression(opstack.Pop));
   end;
   AddOperator(e);
   exptree:=e;
   AcceptToken(tkclosebracket);
   CheckToken(binaryop+endtoken+[tkarrayopen,tkarrayclose,tkclosebracket,tkderef]);
   preoperator:=false;
   funccall:=false;
  end
  else
  begin
   AcceptToken(tkopenbracket);
   ParseExpression_r;
   AcceptToken(tkclosebracket);
   preoperator:=false;
   CheckToken(binaryop+endtoken+[tkarrayclose,tkclosebracket,tkderef]);
  end;
 end;
 tknot:
 begin
   e:=TExpression.create(cgen,nodelist);
   e.info:=ReadTokenInfo;
   e.unary:=true;
   AcceptToken(e.info.token);
   while (opstack.count>0) and CompareOperator(TExpression(opstack.Peek),e) do
   begin
    AddOperator(TExpression(opstack.Pop));
   end;
   opstack.push(e);
   CheckToken(unaryop+[tkident,tkinteger,tkfloat,tkstring,tkopenbracket]);
   preoperator:=true;
  funccall:=false;
 end;
end;
until sc.TokenType in [tksemicolon,tkcomma,tkassignment,tkrange,tkarrayclose,tkclosebracket]+endtoken;

while (opstack.count>0) do
begin
 AddOperator(TExpression(opstack.Pop));
end;
result:=exptree;
opstack.free;
end;

function TParser.ParseExpression:TExpression;
var
 start:integer;
begin
start:=expstack.count;
result:=ParseExpression_r;
if expstack.count-start<>1 then
 raise ESyntaxError.Create('Expression expected but '+sc.TokenString+' found');
expstack.pop;
end;

procedure TParser.AssignExpression(exp1,exp2:TExpression);
begin
if (exp1.typeinfo.BaseType=btsingle) and (exp2.info.Token=tkinteger) then
begin
 exp2.info.Token:=tkfloat;
 exp2.SetExpressionType(btsingle);
end;   
if (exp1.info.Token=tkident) and (exp2.info.Token in [tkinteger,tkfloat]) then
begin
 case exp1.typeinfo.BaseType of
  btinteger,btenum:
  begin
   cgen.StoreIntLocalConst(exp1.symbolinfo,strtoint(exp2.info.TokenString));
   exit;
  end;
  btsingle:
  begin
   cgen.StoreFloatLocalConst(exp1.symbolinfo,strtofloat(exp2.info.TokenString));
   exit;
  end;
  btboolean:
  begin
   cgen.StoreByteLocalConst(exp1.symbolinfo,boolean(strtoint(exp2.info.TokenString)));
  end;
 end;
end;
exp2.GenerateCode;
case exp2.typeinfo.BaseType of
 btinteger,btenum,btclass,btclassref:
 begin
  case exp1.typeinfo.BaseType of
   btinteger,btenum,btclass,btclassref:
   begin
    if exp1.info.Token=tkident then
    begin
     cgen.StoreIntLocal(exp1.symbolinfo);
    end
    else
    begin
     cgen.Mov_EDX_EAX;
     exp1.GenerateLValueCode;
     cgen.StoreIntRefEDI;
    end;
   end;
   btsingle:
   begin
    cgen.Mov_EDX_EAX;
    exp1.GenerateLValueCode;
    cgen.StoreFloatRefIntEDI;
   end;
  end;
 end;
 btboolean:
 begin
  if exp1.info.Token=tkident then
  begin
   cgen.StoreByteLocal(exp1.symbolinfo);
  end
  else
  begin
   cgen.Mov_EDX_EAX;
   exp1.GenerateLValueCode;
   cgen.StoreByteRefEDI;
  end; 
 end;
 btsingle:
 begin
  if exp1.info.Token=tkident then
  begin
   cgen.StoreFloatLocal(exp1.symbolinfo);
  end
  else
  begin
   exp1.GenerateLValueCode;
   cgen.StoreFloatRefEDI;
  end;
 end;
 btrecord,btarray:
 begin
  cgen.Mov_ESI_EDI;
  exp1.GenerateLValueCode;
  cgen.MoveBlock(exp1.typeinfo.Size);
 end;
 btstring:
 begin
  cgen.Mov_ESI_EDI;
  exp1.GenerateLValueCode;
  cgen.CallAddr(@System_StringCopy);
 end; 
end;

end;

procedure TParser.ParseExpressionStatement;
var
 exp1,exp2:TExpression;
// r:TParam;
begin
exp1:=ParseExpression;
CheckToken([tksemicolon,tkassignment]);
case sc.TokenType of
 tksemicolon:
  begin
   exp1:=exp1.ReadExpression;
   if exp1.info.Token<>tkfunccall then
    exp1.SyntaxError('Statement expected but expression of type "'+exp1.typeinfo.Name+'" found');
   exp1.GenerateCode;
   if exp1.typeinfo.BaseType=btsingle then
    cgen.PopFPUReg;
  end;
 tkassignment:
  begin
   exp1:=exp1.ReadExpression;
   if not exp1.IsLValue then
    SyntaxError('Left side cannot be assinged to.');
   AcceptToken(tkassignment);
   exp2:=ParseExpression;
   exp2.expectedtype:=exp1.typeinfo.basetypeptr;
   exp2:=exp2.ReadExpression;
   if not exp2.typeinfo.CanAssignTo(exp1.typeinfo) then
    IncompatibleTypes(exp2.typeinfo,exp1.typeinfo);
    AssignExpression(exp1,exp2);
   //exp2.GenerateCode;
   //cgen.Mov_EDX_EAX;
   //exp1.GenerateLValueCode;
  end;
end;
end;

procedure TParser.ParseIfStatement;
var
 exp:TExpression;
 elselabel,endlabel:integer;
begin
AcceptToken(tkif);
exp:=ParseExpression;
exp:=exp.ReadExpression;
if exp.typeinfo.BaseType<>btboolean then
 SyntaxError('Boolean expression expected');
exp.GenerateCode;
AcceptToken(tkthen);
elselabel:=cgen.JumpFalse;
ParseStatement;
if sc.TokenType=tkelse then
begin
 AcceptToken(tkelse);
 endlabel:=cgen.Jump;
 cgen.SetJumpTarget(elselabel,cgen.Code.Size);
 ParseStatement;
 cgen.SetJumpTarget(endlabel,cgen.Code.Size);
end
else
begin
 cgen.SetJumpTarget(elselabel,cgen.Code.Size);
end;

end;

procedure TParser.ParseWhileStatement;
var
 exp:TExpression;
 i,j,startloop,endloop:integer;
 startbreakjmp,startcontjmp:integer;
begin
inc(loopdepth);
startbreakjmp:=length(breakjmp);
startcontjmp:=length(contjmp);
AcceptToken(tkwhile);
startloop:=cgen.Code.Size;
exp:=ParseExpression;
exp:=exp.ReadExpression;
if exp.typeinfo.BaseType<>btboolean then
 SyntaxError('Boolean expression expected');
exp.GenerateCode;
endloop:=cgen.JumpFalse;
AcceptToken(tkdo);
ParseStatement;
j:=cgen.Jump;
cgen.SetJumpTarget(j,startloop);
cgen.SetJumpTarget(endloop,cgen.code.Size);
dec(loopdepth);
for i:=startbreakjmp to length(breakjmp)-1 do
 cgen.SetJumpTarget(breakjmp[i],cgen.Code.size);
setlength(breakjmp,startbreakjmp);
for i:=startcontjmp to length(contjmp)-1 do
 cgen.SetJumpTarget(contjmp[i],startloop);
setlength(contjmp,startcontjmp);

end;

procedure TParser.ParseForStatement;
var
 exp1,exp2,exp3:TExpression;
 i,j,startloop,endloop,nextloop:Integer;
 startbreakjmp,startcontjmp:integer;
begin
inc(loopdepth);
startbreakjmp:=length(breakjmp);
startcontjmp:=length(contjmp);

AcceptToken(tkfor);
exp1:=ParseExpression;
exp1:=exp1.ReadExpression;
if exp1.info.Token<>tkident then SyntaxError('Local variable expected');
if (exp1.symbolinfo.ParamType<>ptnone) and (exp1.symbolinfo.Flags=[]) then
 SyntaxError('Local variable expected');
if not exp1.typeinfo.IsInteger then SyntaxError('Integer variable expected');
AcceptToken(tkassignment);
exp2:=ParseExpression;
exp2.expectedtype:=exp1.typeinfo.basetypeptr;
exp2:=exp2.ReadExpression;
if not exp2.typeinfo.CanAssignTo(exp1.typeinfo) then
 IncompatibleTypes(exp2.typeinfo,exp1.typeinfo);
AssignExpression(exp1,exp2);
AcceptToken(tkto);

exp3:=ParseExpression;
exp3:=exp3.ReadExpression;
if not exp3.typeinfo.CanAssignTo(exp1.typeinfo) then
 IncompatibleTypes(exp3.typeinfo,exp1.typeinfo);
exp3.GenerateCode;
cgen.SubIntLocal(exp1.symbolinfo);
cgen.AddIntConst(1);
cgen.Mov_EBX_EAX;
cgen.Test_EBX;

startloop:=cgen.Code.Size;
endloop:=cgen.Jump_Zero;
AcceptToken(tkdo);
ParseStatement;

nextloop:=cgen.Code.Size;
cgen.IncVar(exp1.symbolinfo);
cgen.DecEBX;
j:=cgen.Jump;
cgen.SetJumpTarget(j,startloop);
cgen.SetJumpTarget(endloop,cgen.code.Size);

for i:=startbreakjmp to length(breakjmp)-1 do
 cgen.SetJumpTarget(breakjmp[i],cgen.Code.size);
setlength(breakjmp,startbreakjmp);
for i:=startcontjmp to length(contjmp)-1 do
 cgen.SetJumpTarget(contjmp[i],nextloop);
setlength(contjmp,startcontjmp);


dec(cgen.ebx_count);
if cgen.ebx_count>0 then cgen.PopReg(regebx);
dec(loopdepth);


end;

procedure TParser.ParseRepeatStatement;
var
 exp:TExpression;
 i,jmp,startloop:integer;
 startbreakjmp,startcontjmp:integer;
begin
inc(loopdepth);
AcceptToken(tkrepeat);
startbreakjmp:=length(breakjmp);
startcontjmp:=length(contjmp);
startloop:=cgen.Code.Size;
while sc.TokenType<>tkuntil do
begin
 ParseStatement;
 if sc.TokenType<>tkuntil then AcceptToken(tksemicolon);
end;
AcceptToken(tkuntil);

exp:=ParseExpression;                        
exp:=exp.ReadExpression;
if exp.typeinfo.BaseType<>btboolean then
 SyntaxError('Boolean expression expected');
exp.GenerateCode;
jmp:=cgen.JumpFalse;
cgen.SetJumpTarget(jmp,startloop);
dec(loopdepth);
for i:=startbreakjmp to length(breakjmp)-1 do
 cgen.SetJumpTarget(breakjmp[i],cgen.Code.size);
setlength(breakjmp,startbreakjmp);
for i:=startcontjmp to length(contjmp)-1 do
 cgen.SetJumpTarget(contjmp[i],startloop);
setlength(contjmp,startcontjmp);
end;

procedure TParser.ParseExitStatement;
var
 l:integer;
begin
 AcceptToken(tkexit);
 l:=length(exitjmp);
 setlength(exitjmp,l+1);
 exitjmp[l]:=cgen.Jump;
end;

procedure TParser.ParseBreakStatement;
var
 l:integer;
begin
if loopdepth<1 then SyntaxError('Break outside of loop');
AcceptToken(tkbreak);
l:=length(breakjmp);
setlength(breakjmp,l+1);
breakjmp[l]:=cgen.Jump;
end;

procedure TParser.ParseContinueStatement;
var
 l:integer;
begin
if loopdepth<1 then SyntaxError('Continue outside of loop');
AcceptToken(tkcontinue);
l:=length(contjmp);
setlength(contjmp,l+1);
contjmp[l]:=cgen.Jump;
end;

procedure TParser.ParseStatement;
begin
cgen.NewStatement;
CheckToken([tkbegin,tkopenbracket,tkif,tkwhile,tkfor,tkrepeat,tkend,tkbreak,tkexit,tkcontinue]+expressionstart);
case sc.TokenType of
 tkbegin:ParseCodeBlock;
 tkident,tkopenbracket,tkinteger,tkfloat,tkstring,tkadd,tksub,tknot,tkaddress:
  begin
   ParseExpressionStatement;
  end;
 tkif:ParseIfStatement;
 tkwhile:ParseWhileStatement;
 tkfor:ParseForStatement;
 tkrepeat:ParseRepeatStatement;
 tkexit:ParseExitStatement;
 tkbreak:ParseBreakStatement;
 tkcontinue:ParseContinueStatement;
 tksemicolon:AcceptToken(tksemicolon);
 end;
end;


procedure TParser.ParseCodeBlock;
begin
AcceptToken(tkbegin);
while sc.TokenType<>tkend do
begin
ParseStatement;
if sc.TokenType<>tkend then AcceptToken(tksemicolon);
end;
AcceptToken(tkend);

end;

procedure TParser.ParseMethod(mt:TSymbolInfo);
var
 i:integer;
begin
cclass:=mt.Parent;
cgen.BeginFunction(mt);
checktoken([tkvar,tkbegin]);
while sc.TokenType<>tkbegin do
begin
ParseVarBlock;
checktoken([tkvar,tkbegin]);
end;
ParseCodeBlock;
AcceptToken(tksemicolon);
for i:=0 to length(exitjmp)-1 do
 cgen.SetJumpTarget(exitjmp[i],cgen.Code.Size);
exitjmp:=nil; 
cgen.EndFunction;
nodelist.clear;
end;

procedure TParser.ParseImplementationSection;
var
 m:TSymbolInfo;
begin
unitinfo.InterfaceSec:=false;
AcceptToken(tkimplementation);
CheckToken([tkprocedure,tkfunction,tkconstructor,tkdestructor,tkclass,tkend]);
while sc.TokenType<>tkend do
begin
 m:=ParseMethodImpl;
 ParseMethod(m);
end;
AcceptToken(tkend);
AcceptToken(tkpoint);
m:=unitinfo.CheckImplemention;
if m<>nil then
 SyntaxError('Unsatisfield definition: "'+m.TypeInfo.Name+'"');

end;

function TParser.ParseComplexIdent:string;
begin
 result:=sc.TokenString;
 AcceptToken(tkident);
 while (sc.TokenType=tkpoint) do
 begin
  AcceptToken;
  result:=result+'.'+sc.TokenString;
  AcceptToken(tkident);
 end; 
end;

function TParser.Parse:TUnitInfo;
var
 s:string;
begin
AcceptToken(tkunit);
s:=ParseComplexIdent;
AcceptToken(tksemicolon);
unitinfo:=TUnitInfo.create(s);
result:=unitinfo;
ParseInterfaceSection;
ParseImplementationSection;
unitinfo.RegisterClasses;
unitinfo:=nil;
end;

constructor TParser.create(st:TStream);
begin
inherited create;
sc:=TScanner.create(st);
cgen:=TCodeGenerator.create;
nodelist:=TObjectList.Create;
expstack:=TStack.Create;
OnLoadUnit:=LoadUnit;
end;

destructor TParser.destroy;
begin
nodelist.free;
cgen.free;
unitinfo.free;
sc.free;
expstack.free;
inherited;
end;

function CompileUnit(st:TStream):TUnitInfo;
var
 p:TParser;
begin
p:=TParser.create(st);
try
 result:=p.Parse;
finally
 p.Free;
end;
end;

function CompileUnit(const source:string):TUnitInfo;
var
 st:TStream;
begin
st:=TStringStream.Create(source);
try
 result:=CompileUnit(st);
finally
 st.Free;
end;
end;

end.
